/*
 * peap.c  contains the interfaces that are called from eap
 *
 * Version:     $Id: peap.c,v 1.5 2003/09/18 14:03:02 aland Exp $
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *   Copyright 2003 Alan DeKok <aland@freeradius.org>
 */
#include "eap_tls.h"
#include "eap_peap.h"
/*
EAP-TLV format 
    
    0                   1                   2                   3 
    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ 
   |M|R|            TLV Type       |            Length             | 
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ 
   |                              Value... 
   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+ 

*/

/*
 *      Verify that the tunneled packet is valid.
 */

static int tunneledpacket_verify(const uint8_t *data, unsigned int data_len)
{
	eap_packet_t *eap_packet = (eap_packet_t *) data;
	uint8_t eap_type;
	char identity[256];

	if (!data) {
		return 0;
	}

	while (data_len > 0) {
		eap_type = *data;
		data_len--;
		data++;

		if (data_len == 0) {
			return 0;
		}

		if (eap_type == PW_EAP_IDENTITY)  {
			memcpy(identity, data, data_len);
			identity[data_len] = '\0';
			DEBUG2("  rlm_eap_peap: Identity - %s", identity);
			return 1;
		}
		if (eap_type == PW_EAP_RESPONSE) {
			if (eap_packet->data[0] == PW_EAP_TLV) {
				DEBUG2("  rlm_eap_peap: Received EAP-TLV response.");
				return 1;
			}
		}
		if (eap_type == PW_EAP_MSCHAPV2){
			DEBUG2("  rlm_eap_peap: MSCHAPV2");
			return 1;
		}
		return 0;
		}
	return 0;
}

/*
 *      Convert tunneled attributes to the VALUE_PAIR's
 */

static VALUE_PAIR *tunneleddata2vp(EAP_DS *eap_ds, const uint8_t *data, unsigned int data_len)
{
	VALUE_PAIR *vp = NULL;
	eap_packet_t *eap_message;
	int total_length;

	total_length = EAP_HEADER_LEN + data_len;
	eap_message = malloc(total_length);
	if (!eap_message) {
		radlog(L_ERR, "rlm_eap_peap: out of memory");
		return NULL;
	}

	eap_message->code = PW_EAP_RESPONSE;
	eap_message->id = eap_ds->response->id;
	total_length = htons(total_length);
	memcpy(eap_message->length, &total_length, sizeof(uint16_t));
	memcpy(&(eap_message->data[0]), data, data_len);
	
	vp = paircreate(PW_EAP_MESSAGE, PW_TYPE_OCTETS);
	if (!vp) {
		DEBUG2("  rlm_eap_peap: Failure in creating VP");
		return NULL;
	}

	memcpy(vp->strvalue, eap_message, EAP_HEADER_LEN + data_len);
	vp->length = EAP_HEADER_LEN + data_len;

	free(eap_message);

	return vp;
}

/*
 *      Convert VALUE_PAIR's to tunneled attributes, and write them
 *      to an SSL session.
 *
 *      The ONLY VALUE_PAIR's which may be passed to this function
 *      are ones which can go inside of a RADIUS (i.e. diameter)
 *      packet.  So no server-configuration attributes, or the like.
 */

static int vp2tunneleddata(tls_session_t *tls_session, VALUE_PAIR *eap_message)
{
#if 1
	record_plus(&tls_session->clean_in, eap_message->strvalue + EAP_HEADER_LEN, eap_message->length-EAP_HEADER_LEN);
#else
	record_plus(&tls_session->clean_in, eap_message->strvalue, eap_message->length);
#endif
	tls_handshake_send(tls_session);
	record_init(&tls_session->clean_in);

	return 1;
}

/*
 *	Process the "diameter" contents of the tunneled data.
 */
int eappeap_process(EAP_HANDLER *handler, tls_session_t *tls_session)
{
	int err;
	peap_tunnel_t *t = tls_session->opaque;
	REQUEST *fake;
	VALUE_PAIR *vp;
	int rcode = PW_AUTHENTICATION_REJECT;
	const uint8_t *data;
	unsigned int data_len;
	unsigned char buffer[1024];
	eap_packet_t *eap_message ;
	REQUEST *request = handler->request;
	EAP_DS *eap_ds = handler->eap_ds;

	/*
	 *	Grab the dirty data, and copy it to our buffer.
	 *
	 *	I *really* don't like these 'record_t' things...
	 */
	data_len = record_minus(&tls_session->dirty_in, buffer, sizeof(buffer));
	data = buffer;

	/*
	 *	Write the data from the dirty buffer (i.e. packet
	 *	data) into the buffer which we will give to SSL for
	 *	decoding.
	 *
	 *	Some of this code COULD technically go into the TLS
	 *	module, in eaptls_process(), where it returns EAPTLS_OK.
	 *
	 *	Similarly, the writing of data to the SSL context could
	 *	go there, too...
	 */
	BIO_write(tls_session->into_ssl, buffer, data_len);
	record_init(&tls_session->clean_out);

	/*
	 *	Read (and decrypt) the tunneled data from the SSL session,
	 *	and put it into the decrypted data buffer.
	 */
	err = SSL_read(tls_session->ssl, tls_session->clean_out.data,
		       sizeof(tls_session->clean_out.data));
	if (err < 0) {
		/*
		 *	FIXME: Call SSL_get_error() to see what went
		 *	wrong.
		 */
		radlog(L_INFO, "rlm_eap_peap: SSL_read Error");
		return PW_AUTHENTICATION_REJECT;
	}

	/*
	 *	If there's no data, maybe this is an ACK to an
	 *	MS-CHAP2-Success.
	 */	
	if (err == 0) {
		if (t->status >= TLV_SUCCESS) {
			DEBUG2("  PEAP: Got ACK, and the user was already authenticated.");
			return PW_AUTHENTICATION_ACK;
		} /* else no session, no data, die. */


		/*
		 *	FIXME: Call SSL_get_error() to see what went
		 *	wrong.
		 */
		radlog(L_INFO, "rlm_eap_peap: SSL_read Error");
		return PW_AUTHENTICATION_REJECT;
	}
 
	data_len = tls_session->clean_out.used = err;
	data = tls_session->clean_out.data;

	if (!tunneledpacket_verify(data, data_len)) {
		return PW_AUTHENTICATION_REJECT;
	}
	DEBUG2("  rlm_eap_peap: Tunneled data is valid.");

/* Check if the eap response packet has the type set to TLV 
	and its Result-TLV =Success */

	if (t->status >= TLV_SUCCESS) {
	  eap_message = (eap_packet_t *) data;	
	 
	  if ((eap_message->code == PW_EAP_RESPONSE) &&
            	(eap_message->data[0] == PW_EAP_TLV) &&
            	(data[10] == 1)) 
		{
			DEBUG2("  rlm_eap_peap: SUCCESSFUL AUTHENTICATION");
			return PW_AUTHENTICATION_ACK;
		}
		DEBUG2("  rlm_eap_peap: CLIENT SENT RESULT-TLV = FAILURE " );
		return PW_AUTHENTICATION_REJECT;
	}

	fake = request_alloc_fake(request);

	rad_assert(fake->packet->vps == NULL);

	fake->packet->vps = tunneleddata2vp(eap_ds, data, data_len);
	if (!fake->packet->vps) {
		return PW_AUTHENTICATION_REJECT;
	}

#ifndef NDEBUG
	if (debug_flag > 0) {
	  printf("  PEAP: Got tunneled EAP-Message\n");

	  for (vp = fake->packet->vps; vp != NULL; vp = vp->next) {
	    putchar('\t');vp_print(stdout, vp);putchar('\n');
	  }
	}
#endif

	vp = pairmake("EAP-Type", "0", T_OP_EQ);
	vp->lvalue = PW_EAP_MSCHAPV2;
	pairadd(&fake->config_items, vp);

	/*
	 *	Tell the request that it's a fake one.
	 */
	vp = pairmake("Freeradius-Proxied-To", "127.0.0.1", T_OP_EQ);
	if (vp) {
		pairadd(&fake->packet->vps, vp);
	}

	/*
	 *	Update other items in the REQUEST data structure.
	 */
	if (!t->username) {
		if ((data[0] == PW_EAP_IDENTITY) && (data_len > 1)) {
			t->username = pairmake("User-Name", "", T_OP_EQ);
			rad_assert(t->username != NULL);
				
			memcpy(t->username->strvalue, data+1, data_len - 1);
			t->username->length = data_len -1;
			t->username->strvalue[t->username->length] = 0;
			DEBUG2("  PEAP: Got tunneled identity of %s", t->username->strvalue);
		}
	}

	if (t->username) {
		vp = paircopy(t->username);
		pairadd(&fake->packet->vps, vp);
		fake->username = pairfind(fake->packet->vps, PW_USER_NAME);
	}

	/*
	 *	Add the State attribute, too, if it exists.
	 */
	if (t->state) {
		DEBUG2("  PEAP: Adding old state with %02x %02x",
		       t->state->strvalue[0], t->state->strvalue[1]);
		vp = paircopy(t->state);
		if (vp) pairadd(&fake->packet->vps, vp);
	}

#ifndef NDEBUG
	if (debug_flag > 0) {
	  printf("  PEAP: Sending tunneled request\n");

	  for (vp = fake->packet->vps; vp != NULL; vp = vp->next) {
	    putchar('\t');vp_print(stdout, vp);putchar('\n');
	  }
	}
#endif

	/*
	 *	Call authentication recursively, which will
	 *	do PAP, CHAP, MS-CHAP, etc.
	 */
	rad_authenticate(fake);

	/*
	 *	Note that we don't do *anything* with the reply
	 *	attributes.
	 *
	 *	When we finally handle EAP, we will need to save
	 *	the State attribute from the reply, in the 'handle',
	 *	and add it to the next fake request which uses this
	 *	handle.
	 */
#ifndef NDEBUG
	if (debug_flag > 0) {
	  printf("  PEAP: Got tunneled reply RADIUS code %d\n",
		 fake->reply->code);

	  for (vp = fake->reply->vps; vp != NULL; vp = vp->next) {
	    putchar('\t');vp_print(stdout, vp);putchar('\n');
	  }
	}
#endif
	if (fake->reply->code == PW_AUTHENTICATION_ACK) {
		DEBUG2("  PEAP: Tunneled authentication is success.");
		t->status = TLV_SUCCESS;
		eappeap_success(handler, tls_session);
		rcode = PW_ACCESS_CHALLENGE;
	} else if (fake->reply->code == PW_AUTHENTICATION_REJECT) {
		DEBUG2("  PEAP: Tunneled authentication in rejected.");
		t->status = TLV_FAILURE;
		eappeap_failure(handler, tls_session);
		rcode = PW_AUTHENTICATION_REJECT;
	} else if (fake->reply->code == PW_ACCESS_CHALLENGE) {
		DEBUG2("  PEAP: Got tunneled Access-Challenge");

		/*
		 *	Keep the State attribute, if necessary.
		 *
		 *	Get rid of the old State, too.
		 */
		pairfree(&t->state);
		pairmove2(&t->state, &fake->reply->vps, PW_STATE);

		/*
		 *	We should really be a bit smarter about this,
		 *	and move over only those attributes which
		 *	are relevant to the authentication request,
		 *	but that's a lot more work, and this "dumb"
		 *	method works in 99.9% of the situations.
		 */
		vp = NULL;
		pairmove2(&vp, &fake->reply->vps, PW_EAP_MESSAGE);

		/*
		 *	There MUST be a Reply-Message in the challenge,
		 *	which we tunnel back to the client.
		 *
		 *	If there isn't one in the reply VP's, then
		 *	we MUST create one, with an empty string as
		 *	it's value.
		 */
		pairmove2(&vp, &fake->reply->vps, PW_REPLY_MESSAGE);

		/*
		 *	Handle the ACK, by tunneling any necessary reply
		 *	VP's back to the client.
		 */
		if (vp) {
			vp2tunneleddata(tls_session, vp);
			pairfree(&vp);
		}
		rcode = PW_ACCESS_CHALLENGE;

	} else {
		DEBUG2("  PEAP: Rejecting tunneled user");
		rcode = PW_AUTHENTICATION_REJECT;
	}
	
	request_free(&fake);
	
	return rcode;
}

/*
 *	Send EAP-Identity request
 */
int eappeap_request_id(EAP_DS *eap_ds, tls_session_t *tls_session)
{
	eap_packet_t eap_packet;
	uint16_t length;

	eap_packet.code = PW_EAP_REQUEST;
	eap_packet.id = eap_ds->response->id + 1;
	length = EAP_HEADER_LEN + 1;
	length = htons(length);
	memcpy(eap_packet.length, &length, sizeof(uint16_t));
	eap_packet.data[0] = PW_EAP_IDENTITY;

	record_plus(&tls_session->clean_in, (const unsigned char *) &eap_packet, sizeof(eap_packet));

	tls_handshake_send(tls_session);
	record_init(&tls_session->clean_in);
	return 1;
}
/*
 *	Send protected EAP-Failure
 *
 *       Result-TLV = Failure
 */
static int eappeap_failure(EAP_HANDLER *handler, tls_session_t *tls_session)
{
	uint8_t *tlv_packet;
	uint16_t length;

	tlv_packet = malloc(11);
	if (!tlv_packet) {
		radlog(L_ERR, "rlm_eap_peap: out of memory");
		return 0;
	}

	tlv_packet[0] = PW_EAP_REQUEST;
	tlv_packet[1] = handler->eap_ds->response->id +1;
	length = 10;
	length = htons(length);
	memcpy(&tlv_packet[2], &length, sizeof(uint16_t));
	tlv_packet[4] = PW_EAP_TLV;
	tlv_packet[5] = 0x80;
	tlv_packet[6] = 3;
	tlv_packet[7] = 0;
	tlv_packet[8] = 2;
	tlv_packet[9] = 0;
	tlv_packet[10] = 2;

	record_plus(&tls_session->clean_in, tlv_packet, 11);

	/*
	 *	FIXME: Check the return code.
	 */
	tls_handshake_send(tls_session);
	record_init(&tls_session->clean_in);
	free(tlv_packet);

	return 1;
}

/*
 *	Send protected EAP-Success
 *
 *       Result-TLV = Success
 */
static int eappeap_success(EAP_HANDLER *handler, tls_session_t *tls_session)
{
	uint8_t *tlv_packet;
	uint16_t length;

	tlv_packet = malloc(11);
	if (!tlv_packet) {
		radlog(L_ERR, "rlm_eap_peap: out of memory");
		return 0;
	}

	tlv_packet[0] = PW_EAP_REQUEST;
	tlv_packet[1] = handler->eap_ds->response->id +1;
	length = 10;
	length = htons(length);
	memcpy(&tlv_packet[2], &length, sizeof(uint16_t));
	tlv_packet[4] = PW_EAP_TLV;
	tlv_packet[5] = 0x80;
	tlv_packet[6] = 3;
	tlv_packet[7] = 0;
	tlv_packet[8] = 2;
	tlv_packet[9] = 0;
	tlv_packet[10] = 1;

	record_plus(&tls_session->clean_in, tlv_packet, 11);

	/*
	 *	FIXME: Check the return code.
	 */
	tls_handshake_send(tls_session);
	record_init(&tls_session->clean_in);
	free(tlv_packet);

	return 1;
}
